import requests
import functools
import logging

from .util import exceptions
from .util.dict_util import format_keys
from .schemas.extensions import get_wrapped_create_function

logger = logging.getLogger(__name__)


def get_request_args(rspec, test_block_config):
    """Format the test spec given values inthe global config

    Todo:
        Add similar functionality to validate/save $ext functions so input
        can be generated from a function

    Args:
        rspec (dict): Test spec
        test_block_config (dict): Test block config

    Returns:
        dict: Formatted test spec

    Raises:
        BadSchemaError: Tried to pass a body in a GET request
    """

    request_args = {}

    # Ones that are required and are enforced to be present by the schema
    required_in_file = [
        "method",
        "url",
    ]

    optional_in_file = [
        "json",
        "data",
        "params",
        "headers",
    ]

    if "method" not in rspec:
        logger.debug("Using default GET method")
        rspec["method"] = "GET"

    headers = rspec.get("headers")
    if headers:
        if "content-type" not in [h.lower() for h in headers.keys()]:
            rspec["headers"]["content-type"] = "application/json"

    try:
        fspec = format_keys(rspec, test_block_config["variables"])
    except exceptions.MissingFormatError as e:
        logger.error("Key not found in format")
        raise

    def add_request_args(keys, optional):
        for key in keys:
            try:
                request_args[key] = fspec[key]
            except KeyError:
                if optional or (key in request_args):
                    continue

                # This should never happen
                raise

    add_request_args(required_in_file, False)
    add_request_args(optional_in_file, True)

    for key in optional_in_file:
        try:
            func = get_wrapped_create_function(request_args[key].pop("$ext"))
        except KeyError:
            pass
        else:
            request_args[key] = func()

    # TODO
    # requests takes all of these - we need to parse the input to get them
    # "files",
    # "auth",
    # "cookies",

    if request_args["method"] == "GET":
        if any(i in request_args for i in ["json", "data"]):
            raise exceptions.BadSchemaError("Can't add json or urlencoded data to a GET request - use query parameters instead?")

    return request_args


class TRequest:
    def __init__(self, rspec, test_block_config):
        """Prepare request

        Args:
            rspec (dict): test spec
            test_block_config (dict): Any configuration for this the block of
                tests

        Raises:
            UnexpectedKeysError: If some unexpected keys were used in the test
                spec. Only valid keyword args to requests can be passed
        """

        expected = {
            "method",
            "url",
            "headers",
            "files",
            "data",
            "params",
            "auth",
            "cookies",
            "hooks",
            "json",
        }

        keyset = set(rspec)

        if not keyset <= expected:
            unexpected = keyset - expected

            msg = "Unexpected keys {}".format(unexpected)
            logger.error(msg)
            raise exceptions.UnexpectedKeysError(msg)

        request_args = get_request_args(rspec, test_block_config)

        logger.debug("Request args: %s", request_args)

        request_args.update(allow_redirects=False)

        # There is no way using requests to make a prepared request that will
        # not follow redicrects, so instead we have to do this. This also means
        # that we can't have the 'pre-request' hook any more because we don't
        # create a prepared request.
        self._session = requests.Session()
        self._prepared = functools.partial(self._session.request, **request_args)

    def run(self):
        """ Runs the prepared request and times it

        Todo:
            time it

        Returns:
            requests.Response: response object
        """

        return self._prepared()



# https://gist.github.com/joshbode/569627ced3076931b02f

import yaml
import uuid
import os.path

from yaml.reader import Reader
from yaml.scanner import Scanner
from yaml.parser import Parser
from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.resolver import Resolver


def makeuuid(loader, node):
    return str(uuid.uuid4())


class LoaderMeta(type):

    def __new__(metacls, __name__, __bases__, __dict__):
        """Add include constructer to class."""

        # register the include constructor on the class
        cls = super().__new__(metacls, __name__, __bases__, __dict__)
        cls.add_constructor('!include', cls.construct_include)
        cls.add_constructor('!uuid', makeuuid)

        return cls


class RememberComposer(Composer):

    """A composer that doesn't forget anchors across documents
    """

    def compose_document(self):
        # Drop the DOCUMENT-START event.
        self.get_event()

        # Compose the root node.
        node = self.compose_node(None, None)

        # Drop the DOCUMENT-END event.
        self.get_event()

        # If we don't drop the anchors here, then we can keep anchors across
        # documents.
        # self.anchors = {}

        return node


class IncludeLoader(Reader, Scanner, Parser, RememberComposer, SafeConstructor,
   Resolver, metaclass=LoaderMeta):
    """YAML Loader with `!include` constructor."""

    def __init__(self, stream):
        """Initialise Loader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        RememberComposer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)

    def construct_include(self, node):
        """Include file referenced at node."""

        filename = os.path.abspath(os.path.join(
            self._root, self.construct_scalar(node)
        ))
        extension = os.path.splitext(filename)[1].lstrip('.')

        if extension not in ('yaml', 'yml'):
            from tavern.exceptions import BadSchemaError
            raise BadSchemaError("Unknown filetype '{}'".format(filename))

        with open(filename, 'r') as f:
            return yaml.load(f, IncludeLoader)



import logging

import pytest

import yaml

from tavern.core import run_test
from tavern.util.exceptions import TavernException
from tavern.util.loader import IncludeLoader
from tavern.schemas.files import verify_tests


logger = logging.getLogger(__name__)


def pytest_collect_file(parent, path):
    """On collecting files, get any files that end in .tavern.yaml as tavern
    test files

    Todo:
        Change this to .tyaml or something?
    """
    if path.strpath.endswith(".tavern.yaml") and path.basename.startswith("test"):
        return YamlFile(path, parent)


def pytest_addoption(parser):
    """Add an option to pass in a global config file for tavern
    """
    parser.addoption(
        "--tavern-global-cfg",
        required=False,
        help="Global configuration file to include in every test",
    )


class YamlFile(pytest.File):

    """Custom `File` class that loads each test block as a different test
    """

    def collect(self):
        """Load each document in the given input file into a different test

        Yields:
            YamlItem: Essentially an individual pytest 'test object'
        """
        for test_spec in yaml.load_all(self.fspath.open(), Loader=IncludeLoader):
            verify_tests(test_spec)
            yield YamlItem(test_spec["test_name"], self, test_spec, self.fspath)


class YamlItem(pytest.Item):

    """Simple wrapper around new test type that can report errors more
    accurately than the default pytest reporting stuff

    At the time of writing this doesn't print the error very nicely, but it
    should be enough to track down what went wrong

    Attributes:
        path (str): filename that this test came from
        spec (dict): The whole dictionary of the test
    """

    def __init__(self, name, parent, spec, path):
        super().__init__(name, parent)
        self.path = path
        self.spec = spec

    def runtest(self):
        global_cfg = self.config.getoption("tavern_global_cfg") or {}

        with open(global_cfg, "r") as gfileobj:
            contents = yaml.load(gfileobj)

        run_test(self.path, self.spec, contents)

    def repr_failure(self, excinfo):
        """ called when self.runtest() raises an exception.

        Todo:
            This stuff is copied from pytest at the moment - needs a bit of
            modifying so that it shows the yaml and all the reasons the test
            failed rather than a traceback
        """

        if not issubclass(excinfo.type, TavernException):
            return super().repr_failure(excinfo)

        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, 0, "{self.path}::{self.name:s}".format(self=self)



import logging

import yaml

from .util import exceptions
from .util.loader import IncludeLoader
from .util.env_vars import check_env_var_settings
from .request import TRequest
from .response import TResponse
from .printer import log_pass, log_fail

from .schemas.files import verify_tests

logger = logging.getLogger(__name__)


def run_test(in_file, test_spec, global_cfg):
    """Run a single tavern test

    Note that each tavern test can consist of multiple requests (log in,
    create, update, delete, etc).

    The global configuration is copied and used as an initial configuration for
    this test. Any values which are saved from any tests are saved into this
    test block and can be used for formatting in later stages in the test.

    Args:
        in_file (str): filename containing this test
        test_spec (dict): The specification for this test
        global_cfg (dict): Any global configuration for this test

    Raises:
        TavernException: If any of the tests failed
    """
    # Initialise test config for this test with the global configuration before
    # starting

    test_block_config = dict(global_cfg)

    if "variables" not in test_block_config:
        test_block_config["variables"] = {}

    if test_spec.get("includes"):
        for included in test_spec["includes"]:
            if "variables" in included:
                check_env_var_settings(included["variables"])
                test_block_config["variables"].update(included["variables"])

    if not test_spec:
        logger.warning("Empty test block in %s", in_file)
        return

    test_block_name = test_spec["test_name"]

    logger.info("Running test : %s", test_block_name)

    # Run tests in a path in order
    for test in test_spec["stages"]:
        name = test["name"]
        rspec = test["request"]
        expected = test["response"]

        try:
            r = TRequest(rspec, test_block_config)
        except exceptions.MissingFormatError:
            log_fail(test, None, expected)
            raise

        logger.info("Running stage : %s", name)

        response = r.run()

        logger.info("Response: '%s' (%s)", response, response.content.decode("utf8"))

        verifier = TResponse(name, expected, test_block_config)

        try:
            saved = verifier.verify(response)
        except exceptions.TavernException:
            log_fail(test, verifier, expected)
            raise
        else:
            log_pass(test, verifier)
            test_block_config["variables"].update(saved)


def run(in_file, tavern_global_cfg):
    """Run all tests contained in a file

    For each test this makes sure it matches the expected schema, then runs it.
    There currently isn't something like pytest's `-x` flag which exits on first
    failure.

    Args:
        in_file (str): file to run tests on
        tavern_global_cfg (str): file containing Global config for all tests

    Returns:
        bool: Whether ALL tests passed or not
    """

    passed = True

    if tavern_global_cfg:
        with open(tavern_global_cfg, "r") as gfileobj:
            global_cfg = yaml.load(gfileobj)
    else:
        global_cfg = {}

    with open(in_file, "r") as infile:
        # Multiple documents per file => multiple test paths per file
        for test_spec in yaml.load_all(infile, Loader=IncludeLoader):
            try:
                verify_tests(test_spec)
            except exceptions.BadSchemaError:
                passed = False
                continue

            try:
                run_test(in_file, test_spec, global_cfg)
            except exceptions.TestFailError:
                passed = False
                continue

    return passed



class TavernException(Exception):
    pass


class BadSchemaError(TavernException):
    """Schema mismatch
    """
    pass


class TestFailError(TavernException):
    """Test failed somehow
    """
    pass


class UnexpectedKeysError(TavernException):
    """Unexpected keys used in request specification
    """
    pass


class MissingFormatError(TavernException):
    """Tried to use a variable in a format string but it was not in the
    available variables
    """
    pass
